from flask import Flask, jsonify, request
from elasticsearch import Elasticsearch
from datetime import datetime
import flask


es = Elasticsearch([{"host": "localhost", "port": 9200}], http_auth=("elastic", "G-sd5cFKbUch6HbzKNNZ"))
app = Flask(__name__)

documents = ['ADHAR', 'UNCRD', 'PODOC', 'PANCR', 'TWIPC', 'RATCR', 'RVCER', 'VHTAX', 
             'FITCER', 'SSCER', 'LPGSV', 'CRIPC', 'INCER', 'HSCER', 'DRVLC', 'CTCER', 
             'BTCER', 'EPSER', 'PMJAY', 'ASTMT', 'CVIPC', 'EPRAN', 'SPCER', 'HLIPC', 
             'LICER', 'SKCER', 'DMCER', 'RSCER', 'HPCER', 'OSFRM', 'DTCER', 'RSBYC', 
             'SSMGR', 'ROR1B', 'HSMGR', 'IRCER', 'CDCER', 'PECER', 'PSCER', 'EGIPC', 
             'RECER', 'ITCER', 'CMCER', 'OBCER', 'NTCER', 'GOVID', 'PVCER', 'MIIPC', 
             'BHCER', 'TRIPC', 'TDCER', 'RDCER', 'SHCER', 'RMCER', 'OSCER', 'FMCER', 
             'EWCER', 'EBCER', 'CNCMY', 'SVCER', 'LWCER', 'CNDAA', 'ADCRD', 'ITICR',
             'NTMST', 'HMIPC', 'SSYCR', 'ESCER', 'NTMKS', 'DGMST', 'GPFCR', 'LHCER',
             'CHCER', 'CPCER', 'GICER', 'LOCER', 'NTLTR', 'HVCER', 'ISMFC', 'PNCER', 
             'SDCRD', 'EGCER', 'NNCER', 'SICRD', 'CVCER', 'SKMST', 'LCSAG', 'PRIPC', 
             'SMCER', 'HKCER', 'DPICR', 'AGCER', 'RLCER', 'NLCER', 'WWCER', 'MNCER', 
             'OTCER', 'DIPCR', 'LVCER', 'UNCER']

def calculate_age(dob):
    today = datetime.today()
    age = today.year - dob.year - ((today.month, today.day) < (dob.month, dob.day))
    return age

def get_age_category(age):
    if age < 10:
        return 'Under 10'
    elif age < 20:
        return '10-20'
    elif age < 30:
        return '20-30'
    elif age < 40:
        return '30-40'
    elif age < 50:
        return '40-50'
    elif age < 60:
        return '50-60'
    elif age < 70:
        return '60-70'
    elif age < 80:
        return '70-80'
    elif age < 90:
        return '80-90'
    else:
        return 'Over 90'

def create_index():
    index_name = 'test'
    body = {
        'mappings': {
            'properties': {
                'doc_type': {'type': 'keyword'},
                'age': {'type': 'integer'},
                'Date': {'type': 'date'},
                'age_category': {'type': 'keyword'},
                'Timeframe': {'type': 'keyword'}
            }
        }
    }

    if not es.indices.exists(index=index_name):
        response = es.indices.create(index=index_name, body=body)
        if 'acknowledged' in response and response['acknowledged']:
            return True
        else:
            return False
    else:
        return True
    
#COUNTING THE DOCUMENTS WITH DATE AND HOUR
@app.route('/documents/count_date_hour', methods=['GET'])
def count_date_hour():
    index = request.args.get('index', 'test')

    body = {
        "size": 0,
        "aggs": {
            "date_based": {
                "terms": {
                    "field": "Date.keyword"
                },
                "aggs": {
                    "timeframe_based": {
                        "terms": {
                            "field": "Timeframe.keyword"
                        }
                    }
                }
            }
        }
    }

    result = es.search(index=index, body=body)

    if 'aggregations' in result and 'date_based' in result['aggregations']:
        date_buckets = result['aggregations']['date_based']['buckets']
        response = []

        for date_bucket in date_buckets:
            date_key = date_bucket['key']
            timeframe_buckets = date_bucket['timeframe_based']['buckets']

            for timeframe_bucket in timeframe_buckets:
                timeframe_key = timeframe_bucket['key']
                doc_count = timeframe_bucket['doc_count']

                doc_info = {
                    "date": date_key,
                    "timeframe": timeframe_key,
                    "count": doc_count
                }

                response.append(doc_info)

    else:
        response = []

    return {
        "status": 200,
        "response": response
    }


#DOC_TYPE VERIFICATION
@app.route('/documents/<string:doc_type>', methods=['GET'])
def select_document(doc_type):
    doc_type= doc_type.upper()
    if doc_type in documents :
        new_line = '\n'
        return {"Status":200,
                "Response":f"You selected the {doc_type.upper()} document {new_line} Please Enter the Date of birth of format[01022015] Query parameter"}
    else:
        return {"Status":404,
                "Response":"Invalid doc_type"}

#Enter DOB, ALONG WITH DOC_TYPE
@app.route('/documents/<string:doc_type>/<string:dob>', methods=['GET'])
def get_document_by_dob(doc_type, dob):
    doc_type = doc_type.upper()
    if doc_type not in documents:
        return  {"Status":404,
                "Response":"Invalid doc_type"}

    dob = datetime.strptime(dob, '%d%m%Y').date()
    age = calculate_age(dob)
    today = datetime.today().strftime('%d%m%Y')
    age_cal = get_age_category(age)
    current_time = datetime.now()
    n_res = current_time.strftime('%I')
    timeframe = f'{str(n_res)}{current_time.strftime("%p")}'

    result = {'doc_type': doc_type, 'age': age, 'Date': today, 'age_category': age_cal, 'Timeframe': timeframe}

    doc_id = datetime.now().strftime('%Y%m%d_%I%M%S')

    bulk_request = [
        {"index": {"_index": "test", "_id": doc_id}},
        result
    ]

    resp = es.bulk(body=bulk_request)

    if resp['errors']:
        return  {"Status":404,
                "Response":"Failed to insert documents in Elasticsearch"}

    return {
        "Status":200,
        "result":result
        
    }

class ElasticsearchAPI:
    @staticmethod
    def search_document(query, field_name):
        body = {
            "query": {
                "match": {
                    field_name: query
                }
            }
        }
        
        res = es.search(index="test", body=body)
        return res


#SEARCH THE DOCUMETNS INSIDE  ELASTICSEARCH
@app.route('/search_documents', methods=['POST'])
def search_document():
    data = flask.request.json
 
    field_name = data['field_name']
    query = data['query']

    result = ElasticsearchAPI.search_document(query, field_name)

    docs = []

    hits = result['hits']['hits']

    for hit in hits:
        docs.append(hit['_source'])

    

    return {
        "status": 200,
        "documents": docs
    }


#COUNT THE Doc_types ACCORDINGLY ON DATE
@app.route('/documents/count_by_day', methods=['GET'])
def count_documents_by_day():
    index = request.args.get('index', 'test')

    body = {
        'aggs': {
            'date-based': {
                'terms': {
                    'field': 'Date.keyword'
                    
                }
            }
        }
    }

   
    result = es.search(index=index, body=body)

    if 'aggregations' in result and 'date-based' in result['aggregations']:
        date_buckets = result['aggregations']['date-based']['buckets']
        counts_by_day = {bucket['key']: bucket['doc_count'] for bucket in date_buckets}
    else:
        counts_by_day = {}

    return {"Status":200,
            "Response":counts_by_day
            }

#COUNT THE DOCUMENTS BY HOURLEY 
@app.route('/documents/count_by_hour', methods=['GET'])
def count_documents_by_hour():
    index = request.args.get('index', 'test')

    body = {
        'aggs': {
            'Hourley-based': {
                'terms': {
                    'field': 'Timeframe.keyword'
                    
                }
            }
        }
    }

   
    result = es.search(index=index, body=body)

    if 'aggregations' in result and 'Hourley-based' in result['aggregations']:
        date_buckets = result['aggregations']['Hourley-based']['buckets']
        counts_by_hour = {bucket['key']: bucket['doc_count'] for bucket in date_buckets}
       
    else:
        counts_by_hour = {}

    return {"Status":200,
            "Response":counts_by_hour            

            }

#FINAL_JSON A NESTED INFORMATION REGARDING THE DATA
@app.route('/documents/final_json', methods=['GET'])
def json_file():
    index = request.args.get('index', 'test')

    body = {
        "size": 0,
        "aggs": {
            "date-based": {
                "terms": {
                    "field": "Date.keyword"
                },
                "aggs": {
                    "timeframe-based": {
                        "terms": {
                            "field": "Timeframe.keyword"
                        },
                        "aggs": {
                            "doc_type-based": {
                                "terms": {
                                    "field": "doc_type.keyword"
                                },
                                "aggs": {
                                    "age_category-based": {
                                        "terms": {
                                            "field": "age_category.keyword"
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    result = es.search(index=index, body=body)

    if 'aggregations' in result and 'date-based' in result['aggregations']:
        date_buckets = result['aggregations']['date-based']['buckets']
        final_json = {}

        for date_bucket in date_buckets:
            date_key = date_bucket['key']
            #print(date_key)
            timeframe_buckets = date_bucket['timeframe-based']['buckets']
            date_counts = {}

            for timeframe_bucket in timeframe_buckets:
                timeframe_key = timeframe_bucket['key']
                doc_type_buckets = timeframe_bucket['doc_type-based']['buckets']
                timeframe_counts = {}

                for doc_type_bucket in doc_type_buckets:
                    doc_type_key = doc_type_bucket['key']
                    age_category_buckets = doc_type_bucket['age_category-based']['buckets']
                    doc_type_counts = {}

                    for age_category_bucket in age_category_buckets:
                        age_category_key = age_category_bucket['key']
                        doc_count = age_category_bucket['doc_count']
                        doc_type_counts[age_category_key] = doc_count

                    timeframe_counts[doc_type_key] = doc_type_counts

                date_counts[timeframe_key] = timeframe_counts

            final_json[date_key] = date_counts
    else:
        final_json = {}

    return {
        "status": 200,
        "response": final_json
    }

#WEIGHING THE DOCUMENT BY DATE
@app.route('/documents/count_doc_weight', methods=['GET'])
def count_doc_weight():
    index = request.args.get('index', 'test')

    body = {
        "size": 0,
        "aggs": {
            "date-based": {
                "terms": {
                    "field": "Date.keyword"
                },
                "aggs": {
                    "doc_types": {
                        "terms": {
                            "field": "doc_type.keyword"
                        },
                        "aggs": {
                            "age_categories": {
                                "terms": {
                                    "field": "age_category.keyword"
                                },
                                "aggs": {
                                    "weight": {
                                        "bucket_script": {
                                            "buckets_path": {
                                                "count": "_count"
                                            },
                                            "script": "params.count / 1000"
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    result = es.search(index=index, body=body)

    if 'aggregations' in result and 'date-based' in result['aggregations']:
        date_buckets = result['aggregations']['date-based']['buckets']
        response = []

        for date_bucket in date_buckets:
            date_key = date_bucket['key']
            doc_type_buckets = date_bucket['doc_types']['buckets']

            for doc_type_bucket in doc_type_buckets:
                doc_type_key = doc_type_bucket['key']
                age_category_buckets = doc_type_bucket['age_categories']['buckets']

                for age_category_bucket in age_category_buckets:
                    age_category_key = age_category_bucket['key']
                    doc_count = age_category_bucket['doc_count']
                    weight = age_category_bucket['weight']['value']

                    doc_type_age_category = f"{doc_type_key}_{age_category_key}"
                    doc_info = {
                        doc_type_age_category: {
                            "doc_type": doc_type_key,
                            "age_category": age_category_key,
                            "weight": weight,
                            "count": doc_count,
                            "Date":date_key
                        }
                    }
                    response.append(doc_info)

    else:
        response = []

    return {
        "status": 200,
        "response": response
    }



if __name__ == "__main__":    
    if create_index():
        app.run(debug=True, port=8000, host="127.0.0.1")
    else:
        print('Failed to create index in Elasticsearch')



###final output